# This Python file uses the following encoding: utf-8
# This module defines the constants needed for the package

colorcycle = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
              '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']

units = {'GST': '°C', 'Air temperature': '°C',
         'Precipitation': 'mm/day', 'Water production': 'mm/day',
         'SWE': 'mm', 'Snow depth': 'mm',
         'SW': 'W m-2', 'SW direct': 'W m-2', 'SW diffuse': 'W m-2',
         'SW up': 'W m-2', 'SW down': 'W m-2', 'SW net': 'W m-2', 'LW net': 'W m-2'}
