""" Imports functions that could be re-used by a user in a script """
#pylint: disable=line-too-long

# creating and loading pickles
from SuPerSim.pickling import load_all_pickles, get_all_stats

# producing all plots for either 1 site or for comparing 2 sites
from SuPerSim.functions_summary import plot_camparison_two_sites, plot_all

# all the individual plotting functions
# plotting directly from the inputs
from SuPerSim.horizon import plot_visible_skymap_from_horizon_file
from SuPerSim.weights import plot_hist_stat_weights_from_input, plot_hist_valid_sim_all_variables_from_input
from SuPerSim.runningstats import plot_aggregating_distance_temp_all_from_input
from SuPerSim.yearlystats import plot_box_yearly_stat_from_inputs
from SuPerSim.topoheatmap import plot_table_mean_GST_aspect_slope_single_altitude_from_inputs, plot_table_mean_GST_aspect_slope_all_altitudes_from_inputs, plot_table_mean_GST_aspect_slope_all_altitudes_polar_from_inputs, plot_permafrost_all_altitudes_polar_from_inputs
from SuPerSim.percentiles import plot_cdf_GST_from_inputs, plot_heatmap_percentile_GST_from_inputs
from SuPerSim.evolution import plot_GST_bkg_vs_evol_quantile_bins_fit_single_site_from_inputs, plot_mean_bkg_GST_vs_evolution_from_inputs, plot_evolution_snow_cover_melt_out_from_inputs, plot_GST_bkg_vs_evol_quantile_bins_fit_two_sites_from_input
from SuPerSim.model import fit_stat_model_GST_from_inputs
from SuPerSim.yearlystats import plot_yearly_quantiles_atmospheric_from_inputs, plot_yearly_quantiles_sim_from_inputs, plot_yearly_quantiles_side_by_side_sim_from_inputs, plot_yearly_quantiles_side_by_side_atmospheric_from_inputs
from SuPerSim.seasonal import plot_sanity_two_variables_one_year_quantiles_from_inputs, plot_sanity_one_year_quantiles_two_periods_from_inputs, plot_sanity_two_variables_two_sites_one_year_quantiles_side_by_side_from_inputs

# plotting from formatted data that a user can produce
from SuPerSim.horizon import plot_visible_skymap
from SuPerSim.weights import plot_hist_stat_weights, plot_hist_valid_sim_all_variables
from SuPerSim.runningstats import plot_aggregating_distance_temp_all
from SuPerSim.yearlystats import plot_box_yearly_stat
from SuPerSim.topoheatmap import plot_table_mean_GST_aspect_slope_single_altitude, plot_table_mean_GST_aspect_slope_all_altitudes, plot_table_mean_GST_aspect_slope_all_altitudes_polar, plot_permafrost_all_altitudes_polar
from SuPerSim.percentiles import plot_cdf_GST, plot_heatmap_percentile_GST
from SuPerSim.evolution import plot_GST_bkg_vs_evol_quantile_bins_fit_single_site, plot_mean_bkg_GST_vs_evolution, plot_evolution_snow_cover_melt_out, plot_GST_bkg_vs_evol_quantile_bins_fit_two_sites
from SuPerSim.model import fit_stat_model_GST
from SuPerSim.yearlystats import plot_yearly_quantiles, plot_yearly_quantiles_side_by_side
from SuPerSim.seasonal import plot_sanity_two_variables_one_year_quantiles, plot_sanity_one_year_quantiles_two_periods, plot_sanity_two_variables_two_sites_one_year_quantiles_side_by_side
